#define _POSIX_C_SOURCE 200112L

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "list.h"

#define PAGE_SIZE         (4096)

/* zarovna hodnotu x na nasobek mocniny dvou */
#define ALIGN(x, n)       (((x) + ((n) - 1)) & (~((n) - 1)))

/* zarovna hodnotu na velikost stranky */
#define ALIGN_PG_SIZE(x)  ALIGN(x, PAGE_SIZE)

/* zarovna hodnotu x na 4 byty */
#define ALIGN4(x)         ALIGN(x, 4)

/* zarovna hodnotu x na 8 bytu */
#define ALIGN8(x)	  ALIGN(x, 8)

/**
 * struktura popisujici jeden slab (deska/plat)
 */
struct slab {
	struct slab *next;  /* odkazy na dalsi/predchozi slaby */
	struct slab *prev;
	uint16_t obj_cnt;   /* pocet aktivnich objektu v danem slabu */
	uint16_t bmap_size; /* velikost bitmapy v bytech */
};


/* makro vracejici metadata daneho slabu */
#define SLAB_HEADER(page) ((struct slab *) (((unsigned char *) (page) + PAGE_SIZE - sizeof(struct slab))))

/* makro vracejici odkaz na datovou cast slabu */
#define SLAB_DATA(slab)   ((unsigned char *) (((unsigned long) slab) & ~(PAGE_SIZE - 1)))

/* makro vracejici odkaz na bitmapu s pouzivanymi objekty */
#define SLAB_BMAP(slab)   ((unsigned char *) (slab) - (slab)->bmap_size)


/**
 * Struktura popisujici alokator pro jeden typ objektu
 */
struct kmem_cache {
	unsigned int obj_size;      /* velikost objektu zaokrouhlena na slova */
	unsigned int objs_in_slab;  /* pocet objektu v jednom slabu */
	unsigned int bmap_size;     /* velikost bitmapy volnych bloku */
	void *private;              /* pomocna hodnota, kterou lze specifikovat dalsi vlasnosti slabu */

	void (*ctor)(void *, struct kmem_cache *); /* konstruktor */
	void (*dtor)(void *, struct kmem_cache *); /* destruktor */

	struct slab *partial_slabs; /* seznam castecne zaplnenych slabu */
	struct slab *full_slabs;    /* seznam zcela zaplnenych slabu */
	struct slab *free_slabs;    /* seznam volnych slabu */
};

/** alokator slouzici k alokaci dalsich alokatoru */
struct kmem_cache cache_cache;


void *kmem_cache_alloc(struct kmem_cache *cache);
void kmem_cache_free(struct kmem_cache *cache, void *obj);

/** celociselne deleni se zaokrouhlovanim nahoru */
static inline uint32_t ceiling_div(uint32_t dividend, uint32_t divisor)
{
        uint32_t q = dividend / divisor;
        uint32_t m = dividend % divisor;
        if (m != 0) q ++;
        return q;
}

/**
 * Provede odhad kolik objektu se vejde do jednoho slabu.
 * Vraci i velikost bitmapy, pro ulozeni informaci o aktivnich/neaktivnich objektech.
 */
static int cache_capacity(uint32_t page_size, uint32_t aligned_size, uint32_t *bmap_size)
{
	/* odhad poctu objektu -- neni zapocitana bitmapa */
	uint32_t obj_cnt_estimate = (page_size - sizeof(struct slab)) / aligned_size;

	/* vypocte velikost bitmapy */
	*bmap_size = ALIGN4(ceiling_div(obj_cnt_estimate, 8));	
	uint32_t obj_cnt   = (page_size - sizeof(struct slab) - *bmap_size) / aligned_size;
	
	return obj_cnt;	
}

/**
 * Ziska jednu stranku. (Provizorni reseni.)
 */
static void *acquire_page()
{
	void *page;
	posix_memalign(&page, PAGE_SIZE, PAGE_SIZE);
	memset(page, 0x00, PAGE_SIZE);
	return page;
}

/**
 * Vytvori prazdny slab a inicializuje v nem vsechny objekty
 */
static void create_free_slab(struct kmem_cache *cache)
{
	unsigned char *page = acquire_page();
	memset(page, 0x00, PAGE_SIZE);

	struct slab *slab = SLAB_HEADER(page);
	slab->bmap_size = cache->bmap_size;
	LIST_INSERT(cache->free_slabs, slab);

	unsigned char *obj = page;
	if (cache->ctor) {
		for (int i = 0; i < cache->objs_in_slab; i++) {
			cache->ctor(obj, cache);
			obj += cache->obj_size;
		}
	}
}


/**
 * Vytvori alokator pro objekty daneho typu
 */
struct kmem_cache *kmem_cache_create(unsigned int size,
	void *private,
	void (*ctor)(void *, struct kmem_cache *),
	void (*dtor)(void *, struct kmem_cache *)) {

	struct kmem_cache *cache = kmem_cache_alloc(&cache_cache);
	cache->obj_size      = ALIGN8(size);
	cache->objs_in_slab  = cache_capacity(PAGE_SIZE, cache->obj_size, &cache->bmap_size);
	cache->private       = private;
	cache->ctor          = ctor;
	cache->dtor          = dtor;

	cache->partial_slabs = NULL;
	cache->full_slabs    = NULL;
	cache->free_slabs    = NULL;

	return cache;
}

/**
 * Inicializuje slab alokator pro alokaci dalsich alokatoru
 */
void kmem_init()
{
	void *cache_cache_slab_pg = acquire_page();
	struct slab *first_slab   = SLAB_HEADER(cache_cache_slab_pg);

	cache_cache.obj_size      = ALIGN8(sizeof(struct kmem_cache));
	cache_cache.objs_in_slab  = cache_capacity(PAGE_SIZE, cache_cache.obj_size, &cache_cache.bmap_size);
	cache_cache.private       = NULL;
	cache_cache.ctor          = NULL;
	cache_cache.dtor          = NULL;

	cache_cache.partial_slabs = NULL;
	cache_cache.full_slabs    = NULL;
	cache_cache.free_slabs    = NULL;

	memset(cache_cache_slab_pg, 0x00, PAGE_SIZE);
	first_slab->bmap_size = cache_cache.bmap_size;
	LIST_INSERT(cache_cache.free_slabs, first_slab);
}

/**
 * Alokuje ze slabu jeden objekt
 */
void *kmem_cache_alloc(struct kmem_cache *cache)
{
	/* najde castecne zaplneny nebo volny slab */
	struct slab *slab = cache->partial_slabs;
	if (!slab) {
		if (!cache->free_slabs) {
			create_free_slab(cache);
		}
		slab = cache->free_slabs;
		LIST_REMOVE(cache->free_slabs, slab);
		LIST_INSERT(cache->partial_slabs, slab);
	}

	/* najde volny blok a oznaci jej za pouzity */
	int idx = -1;
	unsigned char *slab_bmap = SLAB_BMAP(slab);
	for (int i = 0; i < cache->bmap_size; i++) {
		for (int j = 0; j < 8; j++) {
			if (~slab_bmap[i] & (1 << j)) {
				slab_bmap[i] |= 1 << j;
				idx = i * 8 + j;	
				goto found;
			}
		}
	}
found:
	if (idx == -1) abort();

	/* zvysi pocet zaplneny objektu a pripadne preradi slab do odpovidajiciho seznamu */
	slab->obj_cnt++;
	if (slab->obj_cnt == cache->objs_in_slab) {
		LIST_REMOVE(cache->partial_slabs, slab);
		LIST_INSERT(cache->full_slabs, slab);
	}
	
	return SLAB_DATA(slab) + (idx * cache->obj_size);
}

/**
 * Uvolni objekt v danem slabu
 */
void kmem_cache_free(struct kmem_cache *cache, void *obj)
{
	unsigned long obj_addr   = (unsigned long) obj;
	unsigned long slab_addr  = (unsigned long) SLAB_DATA(obj);
	struct slab *slab        = SLAB_HEADER(SLAB_DATA(obj));

	int idx                  = (obj_addr - slab_addr) / cache->obj_size;

	SLAB_BMAP(slab)[idx / 8] ^= 1 << (idx % 8);

	/* postara se o prerazeni slabu do odpovidajiciho seznamu */
	if (slab->obj_cnt == cache->objs_in_slab) LIST_REMOVE(cache->full_slabs, slab);
	else LIST_REMOVE(cache->partial_slabs, slab);

	slab->obj_cnt --;

	if (slab->obj_cnt == 0) LIST_INSERT(cache->free_slabs, slab);
	else LIST_INSERT(cache->partial_slabs, slab);
}

/**
 * Uvolni volne slaby (postara se o volani destruktoru)
 */
void kmem_cache_shrink(struct kmem_cache *cache)
{
	while (cache->free_slabs) {
		struct slab *slab = cache->free_slabs;
		LIST_REMOVE(cache->free_slabs, slab);

		unsigned char *obj = SLAB_DATA(slab);
		if (cache->dtor) {
			for (int i = 0; i < cache->objs_in_slab; i++) {
				cache->dtor(obj, cache);
				obj += cache->obj_size;
			}
		}

		free(SLAB_DATA(slab));
	}	
}

/*
 * Pro ucely testovani
 */

struct foo {
	int a;
	int b;
	int c;
	char d[111];
};

void foo_ctor(void *obj, struct kmem_cache *cache)
{
	struct foo *o = obj;
	o->a = 10;
	o->b = (int) (cache->private);
}


void foo_dtor(void *obj, struct kmem_cache *cache)
{
	printf("destroyed: %x\n", obj);
}



int main()
{
	kmem_init();
	struct kmem_cache *c1 = kmem_cache_create(sizeof(struct foo), (void *) 42, foo_ctor, foo_dtor);
	for (int i = 0; i < 1000; i++) {
		struct foo *foo = kmem_cache_alloc(c1);
		printf("test:%lx -> %i\n", (unsigned long) foo, foo->b);
		kmem_cache_free(c1, foo);
	}
	kmem_cache_shrink(c1);
	return 0;
}
